# fmt: off

__all__ = [
    'BUILTIN_CLASSES'
]

BUILTIN_CLASSES = [
    int,
    float,
    str,
    bool,
    bytes,
    bytearray,
    type,
    object,
    dict,
    tuple,
    set,
    frozenset,
]
