"""
This a hand-curated typeshed file for the builtin `builtins` module, specifically for functions.
"""

from typing import Any, Callable, Iterable, Iterator

####################################################################################################

__all__ = [
    '_hasattr',
    '_getattr',
    '_setattr',
    '_hash',
    '_isinstance',
    '_issubclass',
    '_iter',
    '_repr',
    '_len',
    '_id',
    '_locals',
    '_globals',
    '_map',
    '_next',
    '_range',
]

####################################################################################################


def _hasattr(obj: Any, attr: str) -> bool: ...
def _getattr(obj: Any, attr: str) -> Any: ...
def _setattr(obj: Any, attr: str, val: Any): ...
def _hash(obj: Any) -> int: ...
def _isinstance(obj: Any, classinfo: type | tuple[type, ...]) -> bool: ...
def _issubclass(cls: Any, classinfo: type | tuple[type, ...]) -> bool: ...
def _iter(obj: Any) -> Iterable: ...
def _repr(obj: Any) -> str: ...
def _len(obj: Any) -> int: ...
def _id(obj: Any) -> int: ...
def _locals() -> dict[str, Any]: ...
def _globals() -> dict[str, Any]: ...
def _map(function: Callable, iterable: Iterable, *iterables: Iterable) -> Iterable: ...
def _next(iterator: Iterator) -> Any: ...
def _range(*args: Any) -> Iterator: ...
